"""Remy Role Updater"""
from sh import git
import os
from os import path
import re
import sh
import sys

GITHASH_RE = re.compile(r'commit (\w+)')


class Roles(object):
    """Update Roles to the most current version"""
    ADD = 'Adding'
    UPDATE = 'Updating'
    ROLES = 'roles'
    REPO_PATH = 'repo'

    def __init__(self, repo):
        if path.exists(self.repo_path):
            os.chdir(self.repo_path)
            git.pull('-f', '-u', 'origin', 'master')
        else:
            git.clone(repo, self.REPO_PATH)
            os.chdir(self.repo_path)

        git.submodule('update', '--init', self.ROLES)
        self._metadata = dict()

    def get_submodule_hash(self, submodule_path):
        os.chdir(submodule_path)
        output = str(git('rev-list', '--max-count=1', '--format=%d', 'HEAD'))
        match = GITHASH_RE.match(output)
        os.chdir(self.repo_path)
        return match.groups()[0]

    @property
    def repo_path(self):
        return path.realpath(path.join(self.workspace_path, self.REPO_PATH))

    def update_roles(self):
        os.chdir('%s/%s' % (self.repo_path, self.ROLES))
        git.pull('-f', '-u', 'origin', 'master')
        os.chdir(self.repo_path)

        message = '%s %s %s %s' % (self.UPDATE, self.ROLES, 'to',
                                   self.get_submodule_hash(self.ROLES))
        git.commit('-m', message, '.gitmodules', self.ROLES)
        sys.stdout.write('Committed %s [%s]\n' % (self.ROLES, message))
        try:
            git.push('origin', 'master')
        except sh.ErrorReturnCode_1:
            pass

    @property
    def workspace_path(self):
        return path.realpath(os.path.abspath(os.environ['WORKSPACE']))



"""
Add Jenkins job

"""
from lxml import etree
import getpass
import requests
import sys
import urllib


TEMPLATE = ('<?xml version="1.0" encoding="utf-8"?><project><actions /><descri'
            'ption>Adds or updates chef cookbooks in chef-repo</description><l'
            'ogRotator class="hudson.tasks.LogRotator"><daysToKeep>-1</daysToK'
            'eep><numToKeep>5</numToKeep><artifactDaysToKeep>-1</artifactDaysT'
            'oKeep><artifactNumToKeep>-1</artifactNumToKeep></logRotator><keep'
            'Dependencies>false</keepDependencies><properties><jenkins.plugins'
            '.hipchat.HipChatNotifier_-HipChatJobProperty plugin="hipchat-plug'
            'in@0.1.0"><room>ROOM-NAME</room><startNotification>false</startNo'
            'tification></jenkins.plugins.hipchat.HipChatNotifier_-HipChatJobP'
            'roperty></properties><scm class="hudson.plugins.git.GitSCM" plugi'
            'n="git@1.4.0"><configVersion>2</configVersion><userRemoteConfigs>'
            '<hudson.plugins.git.UserRemoteConfig><name>cookbook</name><refspe'
            'c></refspec><url>{{COOKBOOK-REPO}}</url></hudson.plugins.git.User'
            'RemoteConfig></userRemoteConfigs><branches><hudson.plugins.git.Br'
            'anchSpec><name>master</name></hudson.plugins.git.BranchSpec></bra'
            'nches><localBranch>master</localBranch><disableSubmodules>false</'
            'disableSubmodules><recursiveSubmodules>false</recursiveSubmodules'
            '><doGenerateSubmoduleConfigurations>false</doGenerateSubmoduleCon'
            'figurations><authorOrCommitter>false</authorOrCommitter><clean>fa'
            'lse</clean><wipeOutWorkspace>true</wipeOutWorkspace><pruneBranch'
            'es>false</pruneBranches><remotePoll>false</remotePoll><ignoreNoti'
            'fyCommit>false</ignoreNotifyCommit><useShallowClone>false</useSha'
            'llowClone><buildChooser class="hudson.plugins.git.util.DefaultBui'
            'ldChooser" /><gitTool>Default</gitTool><submoduleCfg class="list"'
            ' /><relativeTargetDir>cookbook</relativeTargetDir><reference></re'
            'ference><excludedRegions></excludedRegions><excludedUsers></exclu'
            'dedUsers><gitConfigName></gitConfigName><gitConfigEmail></gitConf'
            'igEmail><skipTag>false</skipTag><includedRegions></includedRegion'
            's><scmName></scmName></scm><canRoam>true</canRoam><disabled>false'
            '</disabled><blockBuildWhenDownstreamBuilding>false</blockBuildWhe'
            'nDownstreamBuilding><blockBuildWhenUpstreamBuilding>false</blockB'
            'uildWhenUpstreamBuilding><triggers><com.cloudbees.jenkins.GitHubP'
            'ushTrigger plugin="github@1.7"><spec></spec></com.cloudbees.jenki'
            'ns.GitHubPushTrigger></triggers><concurrentBuild>false</concurren'
            'tBuild><builders><hudson.tasks.Shell><command>remy cookbook {{CHE'
            'F-REPO}}</command></hudson.tasks.Shell></builders><publishers><je'
            'nkins.plugins.hipchat.HipChatNotifier plugin="hipchat-plugin@0.1.'
            '0" /></publishers><buildWrappers /></project>')

class JenkinsJob(object):

    HIPCHAT_PROP = 'jenkins.plugins.hipchat.HipChatNotifier_-HipChatJobProperty'

    JOB_PREFIX = 'chef-repo sync '

    def __init__(self, jenkins_host, cookbook_name, cookbook_url, repo_url,
                 username, hipchat_room):
            self.cookbook_name = cookbook_name
            self.cookbook_url = cookbook_url
            self.hipchat = hipchat_room
            self.jenkins_host = jenkins_host
            self.repo_url = repo_url
            self.username = username

    @property
    def job_name(self):
        return '%s%s' % (self.JOB_PREFIX, self.cookbook_name)

    @property
    def jenkins_url(self):
        print self.job_name
        return 'http://%s/createItem?name=%s' % (self.jenkins_host,
                                                 urllib.quote(self.job_name))

    def get_xml(self):
        value = TEMPLATE.replace('{{COOKBOOK-REPO}}', self.cookbook_url)
        doc = etree.fromstring(value.replace('{{CHEF-REPO}}', self.repo_url))
        if self.hipchat is None:
            props = doc.find('properties')
            hipchat = props.find(self.HIPCHAT_PROP)
            props.remove(hipchat)
            publishers = doc.find('publishers')
            hipchat = publishers.find('jenkins.plugins.hipchat.HipChatNotifier')
            publishers.remove(hipchat)
        else:
            hipchat = doc.find('./properties/%s' % self.HIPCHAT_PROP)
            room = hipchat.find('room')
            room.text = self.hipchat
        return etree.tostring(doc, pretty_print=True)

    def create_job(self):
        password = getpass.getpass('Password for %s: ' % self.username)
        response = requests.post(self.jenkins_url,
                                 headers={'content-type': 'text/xml'},
                                 data=self.get_xml(),
                                 auth=(self.username, password))
        if response.status_code != 200:
            sys.stderr.write('Error adding Jenkins job: %s' % response.content)



"""Remy Cookbook Manager"""
from sh import git
import json
import os
from os import path
import re
import sh
import sys

GITHASH_RE = re.compile(r'commit (\w+)')
COOKBOOK_URL_RE = re.compile(r'cookbook\s+([\w\:\.\/\-\@]+)\s')
METADATA_RE = re.compile(r'^(\s+|)([\w]+)\s+(["\']|)([\w\s,\-\.@\:]+)(["\']|)'
                         r'([\s+\n]|,\s+(["\']|)([\w\s,\.@\:]+)(["\']|)'
                         r'[\s+\n])', re.M)
METADATA_KEYWORDS = ['name', 'maintainer', 'maintainer_email', 'license',
                     'description', 'version', 'recipe']


class CookbookManager(object):
    """The CookbookManager manages a chef-repo's cookbooks using Git submodules
    for each cookbook.

    It is meant to be invoked by the Github Jenkins callback, where it will
    update the chef-repo repository with the cookbook that was updated and
    any included submodules.

    """
    ADD = 'Adding'
    UPDATE = 'Updating'
    COOKBOOK = 'cookbook'
    COOKBOOK_PATH = 'cookbook'
    DEPENDENCY = 'dependency cookbook'
    REPO_PATH = 'repo'

    def __init__(self, repo):
        if path.exists(self.repo_path):
            os.chdir(self.repo_path)
            git.pull('-f', '-u', 'origin', 'master')
        else:
            git.clone(repo, self.REPO_PATH)
            os.chdir(self.repo_path)

        git.submodule('update', '--init')
        self._metadata = dict()

    def add_submodule(self, cookbook_name, submodule_path, cookbook_url):
        os.chdir(self.repo_path)
        sys.stdout.write('Adding cookbook: %s\n' % cookbook_name)
        git.submodule('add', cookbook_url, submodule_path)
        git.submodule('update', '--init')

    def commit_object(self, action, object_type, object_name, path_to_object):
        message = '%s %s %s %s %s' % (action, object_type, object_name,
                                      'at' if action == self.ADD else 'to',
                                      self.get_cookbook_hash(path_to_object))
        git.commit('-m', message, '.gitmodules', path_to_object)
        sys.stdout.write('Committed %s [%s]\n' % (path_to_object, message))

    @property
    def cookbook_dependencies(self):
        dependencies_path = path.join(self.cookbook_path, 'dependencies.json')
        if not path.exists(dependencies_path):
            return {}
        return json.loads(open(dependencies_path).read())

    @property
    def cookbook_name(self):
        return self.cookbook_metadata.get('name').strip()

    @property
    def cookbook_metadata(self):
        if self._metadata:
            return self._metadata
        if not path.exists(self.cookbook_metadata_path):
            return dict()
        output = dict()
        with open(self.cookbook_metadata_path) as handle:
            metadata = handle.read()
            for match in METADATA_RE.finditer(metadata):
                if match.group(2) in METADATA_KEYWORDS:
                    if match.group(2) == 'recipe':
                        output[match.group(2)] = {'name': match.group(4),
                                                  'desc': match.group(8)}
                    else:
                        output[match.group(2)] = match.group(4)
        self._metadata = output
        return self._metadata

    @property
    def cookbook_metadata_path(self):
        return path.realpath(os.path.join(self.cookbook_path, 'metadata.rb'))

    @property
    def cookbook_path(self):
        return path.realpath(path.join(self.workspace_path, self.COOKBOOK_PATH))

    @property
    def cookbook_git_url(self):
        os.chdir(self.cookbook_path)
        match = COOKBOOK_URL_RE.match(str(git.remote('-v')))
        os.chdir(self.repo_path)
        return match.groups()[0]

    @property
    def dependency_cookbooks(self):
        return self.cookbook_dependencies.get('cookbooks', dict())

    def get_cookbook_hash(self, submodule_path):
        os.chdir(submodule_path)
        output = str(git('rev-list', '--max-count=1', '--format=%d', 'HEAD'))
        match = GITHASH_RE.match(output)
        os.chdir(self.repo_path)
        return match.groups()[0]

    def process_cookbook(self):
        if not self.cookbook_name:
            sys.stderr.write('Could not find cookbook name, exiting\n')
            sys.exit(1)

        commits = [self.process_cookbook_submodule(self.cookbook_name,
                                                   self.submodule_path,
                                                   self.cookbook_git_url,
                                                   self.COOKBOOK)]

        for dependency in self.dependency_cookbooks:
            dep_path = 'cookbooks/%s' % dependency
            dep_url = self.dependency_cookbooks[dependency]['url']
            commits.append(self.process_cookbook_submodule(dependency,
                                                           dep_path,
                                                           dep_url,
                                                           self.DEPENDENCY))
        # Push any changes
        if any(commits):
            git.push('origin', 'master')

    def process_cookbook_submodule(self, cookbook_name, submodule_path,
                                   cookbook_url, cookbook_type):
        if cookbook_name is None:
            sys.stdout.write('No work to perform this execution\n')
            return
        if self.repo_has_submodule(submodule_path):
            self.update_submodule(submodule_path)
            action = self.UPDATE
        else:
            self.add_submodule(cookbook_name, submodule_path, cookbook_url)
            action = self.ADD

        try:
            self.commit_object(action, cookbook_type, cookbook_name,
                               submodule_path)
            return True
        except sh.ErrorReturnCode_1:
            return False

    def repo_has_submodule(self, submodule_path):
        return path.exists(submodule_path)

    def update_submodule(self, submodule_path):
        os.chdir(submodule_path)
        git.pull('-f', '-u', 'origin', 'master')
        os.chdir(self.repo_path)

    @property
    def repo_path(self):
        return path.realpath(path.join(self.workspace_path, self.REPO_PATH))

    @property
    def submodule_path(self):
        return 'cookbooks/%s' % self.cookbook_name

    @property
    def workspace_path(self):
        return path.realpath(os.path.abspath(os.environ['WORKSPACE']))



"""
Manage GitHub Hooks

"""
import getpass
import json
import requests
import sys

GITHUB_HOST = 'github.com'


class GitHubHook(object):
    """Installs a Jenkins callback hook in GitHub"""
    def __init__(self, github_host, repo_owner, repo_name, hook_url):
        self._github_host = github_host
        self._repo_owner = repo_owner
        self._repo_name = repo_name
        self._hook_url = hook_url
        pass

    @property
    def _github_url(self):
        return 'https://%s/api/v3' % self._github_host or GITHUB_HOST

    def _post_jenkins_hook(self, username, password):
        payload = {"name": "jenkins",
                   "active": True,
                   "events": ["push"],
                   "config": {"jenkins_hook_url": self._hook_url}}
        url = '%s/repos/%s/%s/hooks' % (self._github_url,
                                        self._repo_owner,
                                        self._repo_name)
        response = requests.post(url,
                                 auth=(username, password),
                                 headers={'Content-type': 'application/json'},
                                 data=json.dumps(payload))
        if response.status_code == 200:
            sys.stdout.write('Callback hook enabled\n')
            return True
        sys.stderr.write('Error from Github (%s): %s\n' %
                         (response.status_code, response.json()['message']))
        return False

    def add_callback_hook(self, username=None):
        if username is None:
            username = self._repo_owner
        password = getpass.getpass('Password for %s: ' % username)
        self._post_jenkins_hook(username, password)



from setuptools import setup
from remy import __version__, __description__
import sys

requirements = ['lxml', 'requests', 'sh']

if sys.version_info < (2, 7, 0):
    requirements.append('argparse')

scripts = ['remy=remy.cli:main']

# Build the path to install the support files

setup(name='remy',
      version=__version__,
      author='Gavin M. Roy',
      author_email='gavinmroy@gmail.com',
      packages=['remy'],
      url='https://github.com/gmr/remy',
      summary=__description__,
      long_description=open('README.md').read(),
      install_requires=requirements,
      license=open('LICENSE').read(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: Utilities'],
      entry_points={'console_scripts': scripts},
      zip_safe=True)


"""
CLI stub for arg parsing and command invocation

"""
import argparse
import os
import pwd
import socket
import sys

from remy import cookbooks
from remy import github
from remy import jenkins
from remy import roles
from remy import __description__
from remy import __version__


def add_cookbook_mgmt_options(parser):
    """Add the cookbook management command and arguments.

    :rtype: argparse.ArgumentParser

    """
    cookbook = parser.add_parser('cookbook', help='Invoke in a Jenkins job to '
                                                  'update a cookbook in '
                                                  'chef-repo')
    cookbook.add_argument('repo', action='store',
                          help='Git URL for chef-repo')
    cookbook.set_defaults(func='process_cookbook')

def add_github_hook_options(parser):
    """Add the github jenkins hook command and arguments.

    :rtype: argparse.ArgumentParser

    """
    cookbook = parser.add_parser('github', help='Install the Jenkins callback '
                                                'hook in a GitHub repository')
    cookbook.add_argument('owner', action='store',
                          help='The owner of the GitHub repo')
    cookbook.add_argument('repo', action='store',
                          help='The GitHub repository name')
    domain = socket.gethostname()
    example = 'jenkins.%s' % domain
    cookbook.add_argument('jenkins_hook_url', action='store',
                          help='The jenkins hook URL. For example %s' % example)

    cookbook.add_argument('-g', '--github-host',
                          action='store',
                          dest='github',
                          default=github.GITHUB_HOST,
                          help='Override github.com for a '
                               'GitHub::Enterprise host')
    cookbook.add_argument('-u', '--username',
                          action='store',
                          dest='username',
                          help='Specify a different username than the repo '
                               'owner')

    cookbook.set_defaults(func='github_hooks')


def add_jenkins_job_options(parser):
    """Add a new job to Jenkins for updating chef-repo

    :rtype: argparse.ArgumentParser

    """
    cookbook = parser.add_parser('jenkins', help='Add a new cookbook job to '
                                                 'Jenkins')
    cookbook.add_argument('jenkins', action='store',
                          help='The jenkins server hostname')
    cookbook.add_argument('name', action='store',
                          help='The cookbook name')
    cookbook.add_argument('cookbook', action='store',
                          help='The cookbook git repository URL')
    cookbook.add_argument('chef_repo', action='store',
                          help='The chef-repo git repository URL')
    cookbook.add_argument('-u', '--username',
                          action='store',
                          dest='username',
                          default=pwd.getpwuid(os.getuid())[0],
                          help='Specify a different username than the repo '
                               'owner')
    cookbook.add_argument('-n', '--hipchat-notification',
                          action='store',
                          dest='hipchat',
                          help='Hipchat room for notifications')
    cookbook.set_defaults(func='new_job')

def add_role_options(parser):
    """Add the role command and arguments.

    :rtype: argparse.ArgumentParser

    """
    cookbook = parser.add_parser('roles', help='Invoke in a Jenkins job to '
                                               'update the roles in chef-repo')
    cookbook.add_argument('repo', action='store',
                          help='Git URL for chef-repo')
    cookbook.set_defaults(func='process_roles')


def argparser():
    """Build the argument parser

    :rtype: argparse.ArgumentParser

    """
    parser = argparse.ArgumentParser(description=__description__)
    sparser = parser.add_subparsers()
    add_cookbook_mgmt_options(sparser)
    add_role_options(sparser)
    add_github_hook_options(sparser)
    add_jenkins_job_options(sparser)
    return parser


def main():
    args = argparser().parse_args()
    if args.func == 'process_cookbook':
        if not os.environ.get('WORKSPACE'):
            sys.stderr.write('This command should be run from a Jenkins job\n')
            sys.exit(1)
        obj = cookbooks.CookbookManager(args.repo)
        obj.process_cookbook()

    elif args.func == 'process_roles':
        if not os.environ.get('WORKSPACE'):
            sys.stderr.write('This command should be run from a Jenkins job\n')
            sys.exit(1)
        obj = roles.Roles(args.repo)
        obj.update_roles()

    elif args.func == 'github_hooks':
        obj = github.GitHubHook(args.github, args.owner, args.repo,
                                args.jenkins_hook_url)
        obj.add_callback_hook(args.username)

    elif args.func == 'new_job':
        obj = jenkins.JenkinsJob(args.jenkins, args.name, args.cookbook,
                                 args.chef_repo, args.username, args.hipchat)
        obj.create_job()


if __name__ == '__main__':
    main()